/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2016 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.workspace.reporting.reminder;

import org.openvpms.archetype.rules.patient.reminder.ReminderArchetypes;
import org.openvpms.archetype.rules.patient.reminder.ReminderEvent;
import org.openvpms.component.business.domain.im.act.Act;
import org.openvpms.component.system.common.exception.OpenVPMSException;
import org.openvpms.web.component.app.Context;
import org.openvpms.web.component.im.print.IMObjectReportPrinter;
import org.openvpms.web.component.im.print.InteractiveIMPrinter;
import org.openvpms.web.component.im.report.ContextDocumentTemplateLocator;
import org.openvpms.web.component.im.report.DocumentTemplateLocator;
import org.openvpms.web.component.print.InteractivePrinter;
import org.openvpms.web.component.print.PrinterListener;
import org.openvpms.web.echo.help.HelpContext;
import org.openvpms.web.resource.i18n.Messages;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;


/**
 * Processor for {@link ReminderEvent.Action#PHONE} and {@link ReminderEvent.Action#LIST} events.
 * Prints all of the reminders to a report.
 *
 * @author Tim Anderson
 */
public class ReminderListProcessor extends AbstractReminderBatchProcessor {

    /**
     * The context.
     */
    private final Context context;

    /**
     * The help context.
     */
    private final HelpContext help;

    /**
     * The communication logger. May be {@code null}
     */
    private final ReminderCommunicationLogger logger;

    /**
     * Constructs an {@code ReminderListProcessor}.
     *
     * @param reminders  the reminders
     * @param statistics the reminder statistics
     * @param context    the context
     * @param help       the help context
     * @param logger     if specified, logs listed reminders
     */
    public ReminderListProcessor(List<List<ReminderEvent>> reminders, Statistics statistics, Context context,
                                 HelpContext help, ReminderCommunicationLogger logger) {
        super(reminders, statistics);
        this.context = context;
        this.help = help;
        this.logger = logger;
    }

    /**
     * The processor title.
     *
     * @return the processor title
     */
    public String getTitle() {
        return Messages.get("reporting.reminder.run.list");
    }

    /**
     * Processes the batch.
     */
    public void process() {
        setStatus(Messages.get("reporting.reminder.list.status.begin"));
        final List<ReminderEvent> reminders = getReminders();
        if (!reminders.isEmpty()) {
            try {
                List<Act> acts = new ArrayList<>();
                for (ReminderEvent event : reminders) {
                    acts.add(event.getReminder());
                }
                DocumentTemplateLocator locator = new ContextDocumentTemplateLocator(ReminderArchetypes.REMINDER,
                                                                                     context);
                IMObjectReportPrinter<Act> printer = new IMObjectReportPrinter<>(acts, locator, context);
                final InteractivePrinter iPrinter = createPrinter(printer);

                iPrinter.setListener(new PrinterListener() {
                    public void printed(String printer) {
                        try {
                            updateReminders();
                            notifyCompleted();
                        } catch (Throwable error) {
                            notifyError(error);
                        }
                    }

                    public void cancelled() {
                        notifyCompleted();
                    }

                    public void skipped() {
                        notifyCompleted();
                    }

                    public void failed(Throwable cause) {
                        notifyError(cause);
                    }
                });
                iPrinter.print();
            } catch (OpenVPMSException exception) {
                notifyError(exception);
            }
        } else {
            notifyCompleted();
        }
    }

    /**
     * Restarts processing.
     */
    public void restart() {
        // no-op
    }

    /**
     * Updates a reminder.
     *
     * @param reminder the reminder to update
     * @param date     the last-sent date
     * @return {@code true} if the reminder was updated
     */
    @Override
    protected boolean updateReminder(ReminderEvent reminder, Date date) {
        boolean updated = super.updateReminder(reminder, date);
        if (updated && logger != null) {
            logger.logList(reminder, getContext().getLocation());
        }
        return updated;
    }

    /**
     * Creates a new interactive printer.
     *
     * @param printer the printer to delegate to
     * @return a new interactive printer
     */
    protected InteractivePrinter createPrinter(IMObjectReportPrinter<Act> printer) {
        return createPrinter(Messages.get("reporting.reminder.list.print.title"), printer, context, help);
    }

    /**
     * Creates a new interactive printer.
     *
     * @param title   the dialog title
     * @param printer the printer to delegate to
     * @param context the context
     * @param help    the help context
     * @return a new interactive printer
     */
    protected InteractivePrinter createPrinter(String title, IMObjectReportPrinter<Act> printer, Context context,
                                               HelpContext help) {
        return new InteractiveIMPrinter<>(title, printer, true, context, help);
    }

    /**
     * Notifies the listener (if any) of processing completion.
     */
    @Override
    protected void notifyCompleted() {
        setStatus(Messages.get("reporting.reminder.list.status.end"));
        super.notifyCompleted();
    }

    /**
     * Invoked if an error occurs processing the batch.
     * <p/>
     * Sets the error message on each reminder, and notifies any listener.
     *
     * @param exception the cause
     */
    @Override
    protected void notifyError(Throwable exception) {
        setStatus(Messages.get("reporting.reminder.list.status.failed"));
        super.notifyError(exception);
    }

    /**
     * Returns the context.
     *
     * @return the context
     */
    protected Context getContext() {
        return context;
    }
}
